"""Manager to read and modify config data in JSON files.
"""
# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.
import errno
import io
import json
import os

from traitlets.config import LoggingConfigurable
from traitlets.traitlets import Unicode


def recursive_update(target, new):
    """Recursively update one dictionary using another.

    None values will delete their keys.
    """
    for k, v in new.items():
        if isinstance(v, dict):
            if k not in target:
                target[k] = {}
            recursive_update(target[k], v)
            if not target[k]:
                # Prune empty subdicts
                del target[k]

        elif v is None:
            target.pop(k, None)

        else:
            target[k] = v


class BaseJSONConfigManager(LoggingConfigurable):
    """General JSON config manager

    Deals with persisting/storing config in a json file
    """

    config_dir = Unicode('.')

    def ensure_config_dir_exists(self):
        try:
            os.makedirs(self.config_dir, 0o755)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise

    def file_name(self, section_name):
        return os.path.join(self.config_dir, section_name+'.json')

    def get(self, section_name):
        """Retrieve the config data for the specified section.

        Returns the data as a dictionary, or an empty dictionary if the file
        doesn't exist.
        """
        filename = self.file_name(section_name)
        if os.path.isfile(filename):
            with io.open(filename, encoding='utf-8') as f:
                return json.load(f)
        else:
            return {}

    def set(self, section_name, data):
        """Store the given config data.
        """
        filename = self.file_name(section_name)
        self.ensure_config_dir_exists()

        f = open(filename, 'w', encoding='utf-8')
        with f:
            json.dump(data, f, indent=2)

    def update(self, section_name, new_data):
        """Modify the config section by recursively updating it with new_data.

        Returns the modified config data as a dictionary.
        """
        data = self.get(section_name)
        recursive_update(data, new_data)
        self.set(section_name, data)
        return data



import inspect
import re
import types


def describe(article, value, name=None, verbose=False, capital=False):
    """Return string that describes a value

    Parameters
    ----------
    article : str or None
        A definite or indefinite article. If the article is
        indefinite (i.e. "a" or "an") the appropriate one
        will be infered. Thus, the arguments of ``describe``
        can themselves represent what the resulting string
        will actually look like. If None, then no article
        will be prepended to the result. For non-articled
        description, values that are instances are treated
        definitely, while classes are handled indefinitely.
    value : any
        The value which will be named.
    name : str or None (default: None)
        Only applies when ``article`` is "the" - this
        ``name`` is a definite reference to the value.
        By default one will be infered from the value's
        type and repr methods.
    verbose : bool (default: False)
        Whether the name should be concise or verbose. When
        possible, verbose names include the module, and/or
        class name where an object was defined.
    capital : bool (default: False)
        Whether the first letter of the article should
        be capitalized or not. By default it is not.

    Examples
    --------
    Indefinite description:

    >>> describe("a", object())
    'an object'
    >>> describe("a", object)
    'an object'
    >>> describe("a", type(object))
    'a type'

    Definite description:

    >>> describe("the", object())
    "the object at '0x10741f1b0'"
    >>> describe("the", object)
    "the type 'object'"
    >>> describe("the", type(object))
    "the type 'type'"

    Definitely named description:

    >>> describe("the", object(), "I made")
    'the object I made'
    >>> describe("the", object, "I will use")
    'the object I will use'
    """
    if isinstance(article, str):
        article = article.lower()

    if not inspect.isclass(value):
        typename = type(value).__name__
    else:
        typename = value.__name__
    if verbose:
        typename = _prefix(value) + typename

    if article == "the" or (article is None and not inspect.isclass(value)):
        if name is not None:
            result = "{} {}".format(typename, name)
            if article is not None:
                return add_article(result, True, capital)
            else:
                return result
        else:
            tick_wrap = False
            if inspect.isclass(value):
                name = value.__name__
            elif isinstance(value, types.FunctionType):
                name = value.__name__
                tick_wrap = True
            elif isinstance(value, types.MethodType):
                name = value.__func__.__name__
                tick_wrap = True
            elif type(value).__repr__ in (object.__repr__, type.__repr__):
                name = "at '%s'" % hex(id(value))
                verbose = False
            else:
                name = repr(value)
                verbose = False
            if verbose:
                name = _prefix(value) + name
            if tick_wrap:
                name = name.join("''")
            return describe(article, value, name=name,
                verbose=verbose, capital=capital)
    elif article in ("a", "an") or article is None:
        if article is None:
            return typename
        return add_article(typename, False, capital)
    else:
        raise ValueError("The 'article' argument should "
            "be 'the', 'a', 'an', or None not %r" % article)

    
def _prefix(value):
    if isinstance(value, types.MethodType):
        name = describe(None, value.__self__, verbose=True) + '.'
    else:
        module = inspect.getmodule(value)
        if module is not None and module.__name__ != "builtins":
            name = module.__name__ + '.'
        else:
            name = ""
    return name


def class_of(value):
    """Returns a string of the value's type with an indefinite article.

    For example 'an Image' or 'a PlotValue'.
    """
    if inspect.isclass(value):
        return add_article(value.__name__)
    else:
        return class_of(type(value))


def add_article(name, definite=False, capital=False):
    """Returns the string with a prepended article.

    The input does not need to begin with a charater.

    Parameters
    ----------
    definite : bool (default: False)
        Whether the article is definite or not.
        Indefinite articles being 'a' and 'an',
        while 'the' is definite.
    capital : bool (default: False)
        Whether the added article should have
        its first letter capitalized or not.
    """
    if definite:
        result = "the " + name
    else:
        first_letters = re.compile(r'[\W_]+').sub('', name)
        if first_letters[:1].lower() in 'aeiou':
            result = 'an ' + name
        else:
            result = 'a ' + name
    if capital:
        return result[0].upper() + result[1:]
    else:
        return result


def repr_type(obj):
    """Return a string representation of a value and its type for readable

    error messages.
    """
    the_type = type(obj)
    msg = '{!r} {!r}'.format(obj, the_type)
    return msg



"""Tests for traitlets.config.configurable"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import logging
from unittest import TestCase

from pytest import mark

from traitlets.config.configurable import (
    Configurable,
    LoggingConfigurable,
    SingletonConfigurable,
)
from traitlets.log import get_logger
from traitlets.traitlets import (
    Integer, Float, Unicode, List, Dict, Set, Enum, FuzzyEnum,
    CaselessStrEnum, _deprecations_shown, validate,
)

from traitlets.config.loader import Config

from ...tests._warnings import expected_warnings

class MyConfigurable(Configurable):
    a = Integer(1, help="The integer a.").tag(config=True)
    b = Float(1.0, help="The integer b.").tag(config=True)
    c = Unicode('no config')


mc_help = u"""MyConfigurable(Configurable) options
------------------------------------
--MyConfigurable.a=<Integer>
    The integer a.
    Default: 1
--MyConfigurable.b=<Float>
    The integer b.
    Default: 1.0"""

mc_help_inst=u"""MyConfigurable(Configurable) options
------------------------------------
--MyConfigurable.a=<Integer>
    The integer a.
    Current: 5
--MyConfigurable.b=<Float>
    The integer b.
    Current: 4.0"""

# On Python 3, the Integer trait is a synonym for Int
mc_help = mc_help.replace("<Integer>", "<Int>")
mc_help_inst = mc_help_inst.replace("<Integer>", "<Int>")

class Foo(Configurable):
    a = Integer(0, help="The integer a.").tag(config=True)
    b = Unicode('nope').tag(config=True)
    flist = List([]).tag(config=True)
    fdict = Dict().tag(config=True)


class Bar(Foo):
    b = Unicode('gotit', help="The string b.").tag(config=False)
    c = Float(help="The string c.").tag(config=True)
    bset = Set([]).tag(config=True, multiplicity='+')
    bset_values = Set([2,1,5]).tag(config=True, multiplicity='+')
    bdict = Dict().tag(config=True, multiplicity='+')
    bdict_values = Dict({1:'a','0':'b',5:'c'}).tag(config=True, multiplicity='+')

foo_help=u"""Foo(Configurable) options
-------------------------
--Foo.a=<Int>
    The integer a.
    Default: 0
--Foo.b=<Unicode>
    Default: 'nope'
--Foo.fdict=<key-1>=<value-1>...
    Default: {}
--Foo.flist=<list-item-1>...
    Default: []"""

bar_help=u"""Bar(Foo) options
----------------
--Bar.a=<Int>
    The integer a.
    Default: 0
--Bar.bdict <key-1>=<value-1>...
    Default: {}
--Bar.bdict_values <key-1>=<value-1>...
    Default: {1: 'a', '0': 'b', 5: 'c'}
--Bar.bset <set-item-1>...
    Default: set()
--Bar.bset_values <set-item-1>...
    Default: {1, 2, 5}
--Bar.c=<Float>
    The string c.
    Default: 0.0
--Bar.fdict=<key-1>=<value-1>...
    Default: {}
--Bar.flist=<list-item-1>...
    Default: []"""


class TestConfigurable(TestCase):

    def test_default(self):
        c1 = Configurable()
        c2 = Configurable(config=c1.config)
        c3 = Configurable(config=c2.config)
        self.assertEqual(c1.config, c2.config)
        self.assertEqual(c2.config, c3.config)

    def test_custom(self):
        config = Config()
        config.foo = 'foo'
        config.bar = 'bar'
        c1 = Configurable(config=config)
        c2 = Configurable(config=c1.config)
        c3 = Configurable(config=c2.config)
        self.assertEqual(c1.config, config)
        self.assertEqual(c2.config, config)
        self.assertEqual(c3.config, config)
        # Test that copies are not made
        self.assertTrue(c1.config is config)
        self.assertTrue(c2.config is config)
        self.assertTrue(c3.config is config)
        self.assertTrue(c1.config is c2.config)
        self.assertTrue(c2.config is c3.config)

    def test_inheritance(self):
        config = Config()
        config.MyConfigurable.a = 2
        config.MyConfigurable.b = 2.0
        c1 = MyConfigurable(config=config)
        c2 = MyConfigurable(config=c1.config)
        self.assertEqual(c1.a, config.MyConfigurable.a)
        self.assertEqual(c1.b, config.MyConfigurable.b)
        self.assertEqual(c2.a, config.MyConfigurable.a)
        self.assertEqual(c2.b, config.MyConfigurable.b)

    def test_parent(self):
        config = Config()
        config.Foo.a = 10
        config.Foo.b = "wow"
        config.Bar.b = 'later'
        config.Bar.c = 100.0
        f = Foo(config=config)
        with expected_warnings(['`b` not recognized']):
            b = Bar(config=f.config)
        self.assertEqual(f.a, 10)
        self.assertEqual(f.b, 'wow')
        self.assertEqual(b.b, 'gotit')
        self.assertEqual(b.c, 100.0)

    def test_override1(self):
        config = Config()
        config.MyConfigurable.a = 2
        config.MyConfigurable.b = 2.0
        c = MyConfigurable(a=3, config=config)
        self.assertEqual(c.a, 3)
        self.assertEqual(c.b, config.MyConfigurable.b)
        self.assertEqual(c.c, 'no config')

    def test_override2(self):
        config = Config()
        config.Foo.a = 1
        config.Bar.b = 'or'  # Up above b is config=False, so this won't do it.
        config.Bar.c = 10.0
        with expected_warnings(['`b` not recognized']):
            c = Bar(config=config)
        self.assertEqual(c.a, config.Foo.a)
        self.assertEqual(c.b, 'gotit')
        self.assertEqual(c.c, config.Bar.c)
        with expected_warnings(['`b` not recognized']):
            c = Bar(a=2, b='and', c=20.0, config=config)
        self.assertEqual(c.a, 2)
        self.assertEqual(c.b, 'and')
        self.assertEqual(c.c, 20.0)

    def test_help(self):
        self.assertEqual(MyConfigurable.class_get_help(), mc_help)
        self.assertEqual(Foo.class_get_help(), foo_help)
        self.assertEqual(Bar.class_get_help(), bar_help)

    def test_help_inst(self):
        inst = MyConfigurable(a=5, b=4)
        self.assertEqual(MyConfigurable.class_get_help(inst), mc_help_inst)

    def test_generated_config_enum_comments(self):
        class MyConf(Configurable):
            an_enum = Enum('Choice1 choice2'.split(),
                           help="Many choices.").tag(config=True)

        help_str = "Many choices."
        enum_choices_str = "Choices: any of ['Choice1', 'choice2']"
        rst_choices_str = "MyConf.an_enum : any of ``'Choice1'``|``'choice2'``"
        or_none_str = "or None"

        cls_help = MyConf.class_get_help()

        self.assertIn(help_str, cls_help)
        self.assertIn(enum_choices_str, cls_help)
        self.assertNotIn(or_none_str, cls_help)

        cls_cfg = MyConf.class_config_section()

        self.assertIn(help_str, cls_cfg)
        self.assertIn(enum_choices_str, cls_cfg)
        self.assertNotIn(or_none_str, cls_help)
        ## Check order of Help-msg <--> Choices sections
        self.assertGreater(cls_cfg.index(enum_choices_str),
                           cls_cfg.index(help_str))

        rst_help = MyConf.class_config_rst_doc()

        self.assertIn(help_str, rst_help)
        self.assertIn(rst_choices_str, rst_help)
        self.assertNotIn(or_none_str, rst_help)

        class MyConf2(Configurable):
            an_enum = Enum('Choice1 choice2'.split(),
                           allow_none=True,
                           default_value='choice2',
                           help="Many choices.").tag(config=True)

        defaults_str = "Default: 'choice2'"

        cls2_msg = MyConf2.class_get_help()

        self.assertIn(help_str, cls2_msg)
        self.assertIn(enum_choices_str, cls2_msg)
        self.assertIn(or_none_str, cls2_msg)
        self.assertIn(defaults_str, cls2_msg)
        ## Check order of Default <--> Choices sections
        self.assertGreater(cls2_msg.index(defaults_str),
                           cls2_msg.index(enum_choices_str))

        cls2_cfg = MyConf2.class_config_section()

        self.assertIn(help_str, cls2_cfg)
        self.assertIn(enum_choices_str, cls2_cfg)
        self.assertIn(or_none_str, cls2_cfg)
        self.assertIn(defaults_str, cls2_cfg)
        ## Check order of Default <--> Choices sections
        self.assertGreater(cls2_cfg.index(defaults_str),
                           cls2_cfg.index(enum_choices_str))

    def test_generated_config_strenum_comments(self):
        help_str = "Many choices."
        defaults_str = "Default: 'choice2'"
        or_none_str = "or None"

        class MyConf3(Configurable):
            an_enum = CaselessStrEnum('Choice1 choice2'.split(),
                                      allow_none=True,
                                      default_value='choice2',
                                      help="Many choices.").tag(config=True)

        enum_choices_str = ("Choices: any of ['Choice1', 'choice2'] "
                            "(case-insensitive)")

        cls3_msg = MyConf3.class_get_help()

        self.assertIn(help_str, cls3_msg)
        self.assertIn(enum_choices_str, cls3_msg)
        self.assertIn(or_none_str, cls3_msg)
        self.assertIn(defaults_str, cls3_msg)
        ## Check order of Default <--> Choices sections
        self.assertGreater(cls3_msg.index(defaults_str),
                           cls3_msg.index(enum_choices_str))

        cls3_cfg = MyConf3.class_config_section()

        self.assertIn(help_str, cls3_cfg)
        self.assertIn(enum_choices_str, cls3_cfg)
        self.assertIn(or_none_str, cls3_cfg)
        self.assertIn(defaults_str, cls3_cfg)
        ## Check order of Default <--> Choices sections
        self.assertGreater(cls3_cfg.index(defaults_str),
                           cls3_cfg.index(enum_choices_str))

        class MyConf4(Configurable):
            an_enum = FuzzyEnum('Choice1 choice2'.split(),
                                allow_none=True,
                                default_value='choice2',
                                help="Many choices.").tag(config=True)

        enum_choices_str = ("Choices: any case-insensitive prefix "
                            "of ['Choice1', 'choice2']")

        cls4_msg = MyConf4.class_get_help()

        self.assertIn(help_str, cls4_msg)
        self.assertIn(enum_choices_str, cls4_msg)
        self.assertIn(or_none_str, cls4_msg)
        self.assertIn(defaults_str, cls4_msg)
        ## Check order of Default <--> Choices sections
        self.assertGreater(cls4_msg.index(defaults_str),
                           cls4_msg.index(enum_choices_str))

        cls4_cfg = MyConf4.class_config_section()

        self.assertIn(help_str, cls4_cfg)
        self.assertIn(enum_choices_str, cls4_cfg)
        self.assertIn(or_none_str, cls4_cfg)
        self.assertIn(defaults_str, cls4_cfg)
        ## Check order of Default <--> Choices sections
        self.assertGreater(cls4_cfg.index(defaults_str),
                           cls4_cfg.index(enum_choices_str))



class TestSingletonConfigurable(TestCase):

    def test_instance(self):
        class Foo(SingletonConfigurable): pass
        self.assertEqual(Foo.initialized(), False)
        foo = Foo.instance()
        self.assertEqual(Foo.initialized(), True)
        self.assertEqual(foo, Foo.instance())
        self.assertEqual(SingletonConfigurable._instance, None)

    def test_inheritance(self):
        class Bar(SingletonConfigurable): pass
        class Bam(Bar): pass
        self.assertEqual(Bar.initialized(), False)
        self.assertEqual(Bam.initialized(), False)
        bam = Bam.instance()
        self.assertEqual(Bar.initialized(), True)
        self.assertEqual(Bam.initialized(), True)
        self.assertEqual(bam, Bam._instance)
        self.assertEqual(bam, Bar._instance)
        self.assertEqual(SingletonConfigurable._instance, None)


class TestLoggingConfigurable(TestCase):

    def test_parent_logger(self):
        class Parent(LoggingConfigurable): pass
        class Child(LoggingConfigurable): pass
        log = get_logger().getChild("TestLoggingConfigurable")

        parent = Parent(log=log)
        child = Child(parent=parent)
        self.assertEqual(parent.log, log)
        self.assertEqual(child.log, log)

        parent = Parent()
        child = Child(parent=parent, log=log)
        self.assertEqual(parent.log, get_logger())
        self.assertEqual(child.log, log)

    def test_parent_not_logging_configurable(self):
        class Parent(Configurable): pass
        class Child(LoggingConfigurable): pass
        parent = Parent()
        child = Child(parent=parent)
        self.assertEqual(child.log, get_logger())


class MyParent(Configurable):
    pass

class MyParent2(MyParent):
    pass

class TestParentConfigurable(TestCase):

    def test_parent_config(self):
        cfg = Config({
            'MyParent' : {
                'MyConfigurable' : {
                    'b' : 2.0,
                }
            }
        })
        parent = MyParent(config=cfg)
        myc = MyConfigurable(parent=parent)
        self.assertEqual(myc.b, parent.config.MyParent.MyConfigurable.b)

    def test_parent_inheritance(self):
        cfg = Config({
            'MyParent' : {
                'MyConfigurable' : {
                    'b' : 2.0,
                }
            }
        })
        parent = MyParent2(config=cfg)
        myc = MyConfigurable(parent=parent)
        self.assertEqual(myc.b, parent.config.MyParent.MyConfigurable.b)

    def test_multi_parent(self):
        cfg = Config({
            'MyParent2' : {
                'MyParent' : {
                    'MyConfigurable' : {
                        'b' : 2.0,
                    }
                },
                # this one shouldn't count
                'MyConfigurable' : {
                    'b' : 3.0,
                },
            }
        })
        parent2 = MyParent2(config=cfg)
        parent = MyParent(parent=parent2)
        myc = MyConfigurable(parent=parent)
        self.assertEqual(myc.b, parent.config.MyParent2.MyParent.MyConfigurable.b)

    def test_parent_priority(self):
        cfg = Config({
            'MyConfigurable' : {
                'b' : 2.0,
            },
            'MyParent' : {
                'MyConfigurable' : {
                    'b' : 3.0,
                }
            },
            'MyParent2' : {
                'MyConfigurable' : {
                    'b' : 4.0,
                }
            }
        })
        parent = MyParent2(config=cfg)
        myc = MyConfigurable(parent=parent)
        self.assertEqual(myc.b, parent.config.MyParent2.MyConfigurable.b)

    def test_multi_parent_priority(self):
        cfg = Config({
            'MyConfigurable': {
                'b': 2.0,
            },
            'MyParent': {
                'MyConfigurable': {
                    'b': 3.0,
                },
            },
            'MyParent2': {
                'MyConfigurable': {
                    'b': 4.0,
                },
                'MyParent': {
                    'MyConfigurable': {
                        'b': 5.0,
                    },
                },
            },
        })
        parent2 = MyParent2(config=cfg)
        parent = MyParent2(parent=parent2)
        myc = MyConfigurable(parent=parent)
        self.assertEqual(myc.b, parent.config.MyParent2.MyParent.MyConfigurable.b)

class Containers(Configurable):
    lis = List().tag(config=True)
    def _lis_default(self):
        return [-1]

    s = Set().tag(config=True)
    def _s_default(self):
        return {'a'}

    d = Dict().tag(config=True)
    def _d_default(self):
        return {'a' : 'b'}

class TestConfigContainers(TestCase):
    def test_extend(self):
        c = Config()
        c.Containers.lis.extend(list(range(5)))
        obj = Containers(config=c)
        self.assertEqual(obj.lis, list(range(-1,5)))

    def test_insert(self):
        c = Config()
        c.Containers.lis.insert(0, 'a')
        c.Containers.lis.insert(1, 'b')
        obj = Containers(config=c)
        self.assertEqual(obj.lis, ['a', 'b', -1])

    def test_prepend(self):
        c = Config()
        c.Containers.lis.prepend([1,2])
        c.Containers.lis.prepend([2,3])
        obj = Containers(config=c)
        self.assertEqual(obj.lis, [2,3,1,2,-1])

    def test_prepend_extend(self):
        c = Config()
        c.Containers.lis.prepend([1,2])
        c.Containers.lis.extend([2,3])
        obj = Containers(config=c)
        self.assertEqual(obj.lis, [1,2,-1,2,3])

    def test_append_extend(self):
        c = Config()
        c.Containers.lis.append([1,2])
        c.Containers.lis.extend([2,3])
        obj = Containers(config=c)
        self.assertEqual(obj.lis, [-1,[1,2],2,3])

    def test_extend_append(self):
        c = Config()
        c.Containers.lis.extend([2,3])
        c.Containers.lis.append([1,2])
        obj = Containers(config=c)
        self.assertEqual(obj.lis, [-1,2,3,[1,2]])

    def test_insert_extend(self):
        c = Config()
        c.Containers.lis.insert(0, 1)
        c.Containers.lis.extend([2,3])
        obj = Containers(config=c)
        self.assertEqual(obj.lis, [1,-1,2,3])

    def test_set_update(self):
        c = Config()
        c.Containers.s.update({0,1,2})
        c.Containers.s.update({3})
        obj = Containers(config=c)
        self.assertEqual(obj.s, {'a', 0, 1, 2, 3})

    def test_dict_update(self):
        c = Config()
        c.Containers.d.update({'c' : 'd'})
        c.Containers.d.update({'e' : 'f'})
        obj = Containers(config=c)
        self.assertEqual(obj.d, {'a':'b', 'c':'d', 'e':'f'})

    def test_update_twice(self):
        c = Config()
        c.MyConfigurable.a = 5
        m = MyConfigurable(config=c)
        self.assertEqual(m.a, 5)

        c2 = Config()
        c2.MyConfigurable.a = 10
        m.update_config(c2)
        self.assertEqual(m.a, 10)

        c2.MyConfigurable.a = 15
        m.update_config(c2)
        self.assertEqual(m.a, 15)

    def test_update_self(self):
        """update_config with same config object still triggers config_changed"""
        c = Config()
        c.MyConfigurable.a = 5
        m = MyConfigurable(config=c)
        self.assertEqual(m.a, 5)
        c.MyConfigurable.a = 10
        m.update_config(c)
        self.assertEqual(m.a, 10)

    def test_config_default(self):
        class SomeSingleton(SingletonConfigurable):
            pass

        class DefaultConfigurable(Configurable):
            a = Integer().tag(config=True)
            def _config_default(self):
                if SomeSingleton.initialized():
                    return SomeSingleton.instance().config
                return Config()

        c = Config()
        c.DefaultConfigurable.a = 5

        d1 = DefaultConfigurable()
        self.assertEqual(d1.a, 0)

        single = SomeSingleton.instance(config=c)

        d2 = DefaultConfigurable()
        self.assertIs(d2.config, single.config)
        self.assertEqual(d2.a, 5)

    def test_config_default_deprecated(self):
        """Make sure configurables work even with the deprecations in traitlets"""
        class SomeSingleton(SingletonConfigurable):
            pass

        # reset deprecation limiter
        _deprecations_shown.clear()
        with expected_warnings([]):
            class DefaultConfigurable(Configurable):
                a = Integer(config=True)
                def _config_default(self):
                    if SomeSingleton.initialized():
                        return SomeSingleton.instance().config
                    return Config()

        c = Config()
        c.DefaultConfigurable.a = 5

        d1 = DefaultConfigurable()
        self.assertEqual(d1.a, 0)

        single = SomeSingleton.instance(config=c)

        d2 = DefaultConfigurable()
        self.assertIs(d2.config, single.config)
        self.assertEqual(d2.a, 5)

    def test_kwarg_config_priority(self):
        # a, c set in kwargs
        # a, b set in config
        # verify that:
        # - kwargs are set before config
        # - kwargs have priority over config
        class A(Configurable):
            a = Unicode('default', config=True)
            b = Unicode('default', config=True)
            c = Unicode('default', config=True)
            c_during_config = Unicode('never')
            @validate('b')
            def _record_c(self, proposal):
                # setting b from config records c's value at the time
                self.c_during_config = self.c
                return proposal.value

        cfg = Config()
        cfg.A.a = 'a-config'
        cfg.A.b = 'b-config'
        obj = A(a='a-kwarg', c='c-kwarg', config=cfg)
        assert obj.a == 'a-kwarg'
        assert obj.b == 'b-config'
        assert obj.c == 'c-kwarg'
        assert obj.c_during_config == 'c-kwarg'


class TestLogger(TestCase):

    class A(LoggingConfigurable):
            foo = Integer(config=True)
            bar = Integer(config=True)
            baz = Integer(config=True)

    @mark.skipif(not hasattr(TestCase, 'assertLogs'), reason='requires TestCase.assertLogs')
    def test_warn_match(self):
        logger = logging.getLogger('test_warn_match')
        cfg = Config({'A': {'bat': 5}})
        with self.assertLogs(logger, logging.WARNING) as captured:
            TestLogger.A(config=cfg, log=logger)

        output = '\n'.join(captured.output)
        self.assertIn('Did you mean one of: `bar, baz`?', output)
        self.assertIn('Config option `bat` not recognized by `A`.', output)

        cfg = Config({'A': {'fool': 5}})
        with self.assertLogs(logger, logging.WARNING) as captured:
            TestLogger.A(config=cfg, log=logger)

        output = '\n'.join(captured.output)
        self.assertIn('Config option `fool` not recognized by `A`.', output)
        self.assertIn('Did you mean `foo`?', output)

        cfg = Config({'A': {'totally_wrong': 5}})
        with self.assertLogs(logger, logging.WARNING) as captured:
            TestLogger.A(config=cfg, log=logger)

        output = '\n'.join(captured.output)
        self.assertIn('Config option `totally_wrong` not recognized by `A`.', output)
        self.assertNotIn('Did you mean', output)




from unittest import TestCase

from inspect import Signature, Parameter, signature

from ...traitlets import HasTraits, Int, Unicode

from ..decorators import signature_has_traits


class TestExpandSignature(TestCase):

    def test_no_init(self):
        @signature_has_traits
        class Foo(HasTraits):
            number1 = Int()
            number2 = Int()
            value = Unicode('Hello')

        parameters = signature(Foo).parameters
        parameter_names = list(parameters)

        self.assertIs(parameters['args'].kind, Parameter.VAR_POSITIONAL)
        self.assertEqual('args', parameter_names[0])

        self.assertIs(parameters['number1'].kind, Parameter.KEYWORD_ONLY)
        self.assertIs(parameters['number2'].kind, Parameter.KEYWORD_ONLY)
        self.assertIs(parameters['value'].kind, Parameter.KEYWORD_ONLY)

        self.assertIs(parameters['kwargs'].kind, Parameter.VAR_KEYWORD)
        self.assertEqual('kwargs', parameter_names[-1])

        f = Foo(number1=32, value='World')
        self.assertEqual(f.number1, 32)
        self.assertEqual(f.number2, 0)
        self.assertEqual(f.value, 'World')

    def test_partial_init(self):
        @signature_has_traits
        class Foo(HasTraits):
            number1 = Int()
            number2 = Int()
            value = Unicode('Hello')

            def __init__(self, arg1, **kwargs):
                self.arg1 = arg1

                super(Foo, self).__init__(**kwargs)

        parameters = signature(Foo).parameters
        parameter_names = list(parameters)

        self.assertIs(parameters['arg1'].kind, Parameter.POSITIONAL_OR_KEYWORD)
        self.assertEqual('arg1', parameter_names[0])

        self.assertIs(parameters['number1'].kind, Parameter.KEYWORD_ONLY)
        self.assertIs(parameters['number2'].kind, Parameter.KEYWORD_ONLY)
        self.assertIs(parameters['value'].kind, Parameter.KEYWORD_ONLY)

        self.assertIs(parameters['kwargs'].kind, Parameter.VAR_KEYWORD)
        self.assertEqual('kwargs', parameter_names[-1])

        f = Foo(1, number1=32, value='World')
        self.assertEqual(f.arg1, 1)
        self.assertEqual(f.number1, 32)
        self.assertEqual(f.number2, 0)
        self.assertEqual(f.value, 'World')

    def test_duplicate_init(self):
        @signature_has_traits
        class Foo(HasTraits):
            number1 = Int()
            number2 = Int()

            def __init__(self, number1, **kwargs):
                self.test = number1

                super(Foo, self).__init__(number1=number1, **kwargs)

        parameters = signature(Foo).parameters
        parameter_names = list(parameters)

        self.assertListEqual(parameter_names, ['number1', 'number2', 'kwargs'])

        f = Foo(number1=32, number2=36)
        self.assertEqual(f.test, 32)
        self.assertEqual(f.number1, 32)
        self.assertEqual(f.number2, 36)

    def test_full_init(self):
        @signature_has_traits
        class Foo(HasTraits):
            number1 = Int()
            number2 = Int()
            value = Unicode('Hello')

            def __init__(self, arg1, arg2=None, *pos_args, **kw_args):
                self.arg1 = arg1
                self.arg2 = arg2
                self.pos_args = pos_args
                self.kw_args = kw_args

                super(Foo, self).__init__(*pos_args, **kw_args)

        parameters = signature(Foo).parameters
        parameter_names = list(parameters)

        self.assertIs(parameters['arg1'].kind, Parameter.POSITIONAL_OR_KEYWORD)
        self.assertEqual('arg1', parameter_names[0])

        self.assertIs(parameters['arg2'].kind, Parameter.POSITIONAL_OR_KEYWORD)
        self.assertEqual('arg2', parameter_names[1])

        self.assertIs(parameters['pos_args'].kind, Parameter.VAR_POSITIONAL)
        self.assertEqual('pos_args', parameter_names[2])

        self.assertIs(parameters['number1'].kind, Parameter.KEYWORD_ONLY)
        self.assertIs(parameters['number2'].kind, Parameter.KEYWORD_ONLY)
        self.assertIs(parameters['value'].kind, Parameter.KEYWORD_ONLY)

        self.assertIs(parameters['kw_args'].kind, Parameter.VAR_KEYWORD)
        self.assertEqual('kw_args', parameter_names[-1])

        f = Foo(1, 3, 45, 'hey', number1=32, value='World')
        self.assertEqual(f.arg1, 1)
        self.assertEqual(f.arg2, 3)
        self.assertTupleEqual(f.pos_args, (45, 'hey'))
        self.assertEqual(f.number1, 32)
        self.assertEqual(f.number2, 0)
        self.assertEqual(f.value, 'World')

    def test_no_kwargs(self):
        with self.assertRaises(RuntimeError):
            @signature_has_traits
            class Foo(HasTraits):
                number1 = Int()
                number2 = Int()

                def __init__(self, arg1, arg2=None):
                    pass



#!/usr/bin/env python

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.


# the name of the project
name = 'traitlets'

#-----------------------------------------------------------------------------
# Minimal Python version sanity check
#-----------------------------------------------------------------------------

import sys

v = sys.version_info
if v[:2] < (3,7):
    error = "ERROR: %s requires Python version 3.7 or above." % name
    print(error, file=sys.stderr)
    sys.exit(1)

#-----------------------------------------------------------------------------
# get on with it
#-----------------------------------------------------------------------------

import os
from glob import glob

from distutils.core import setup

pjoin = os.path.join
here = os.path.abspath(os.path.dirname(__file__))
pkg_root = pjoin(here, name)

packages = []
for d, _, _ in os.walk(pjoin(here, name)):
    if os.path.exists(pjoin(d, '__init__.py')):
        packages.append(d[len(here)+1:].replace(os.path.sep, '.'))

version_ns = {}
with open(pjoin(here, name, '_version.py')) as f:
    exec(f.read(), {}, version_ns)


setup_args = dict(
    name            = name,
    version         = version_ns['__version__'],
    scripts         = glob(pjoin('scripts', '*')),
    packages        = packages,
    description     = "Traitlets Python config system",
    long_description= "A configuration system for Python applications.",
    author          = 'IPython Development Team',
    author_email    = 'ipython-dev@scipy.org',
    url             = 'https://github.com/ipython/traitlets',
    license         = 'BSD',
    platforms       = "Linux, Mac OS X, Windows",
    keywords        = ['Interactive', 'Interpreter', 'Shell', 'Web'],
    python_requires = '>=3.7',
    classifiers     = [
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ],
)

if 'develop' in sys.argv or any(a.startswith('bdist') for a in sys.argv):
    import setuptools

setuptools_args = {}

install_requires = setuptools_args['install_requires'] = [
    'ipython_genutils',
]

extras_require = setuptools_args['extras_require'] = {
    'test': ['pytest'],
}

if 'setuptools' in sys.modules:
    setup_args.update(setuptools_args)

if __name__ == '__main__':
    setup(**setup_args)
